#ifndef ENGINE_CLIENT_GRAPHICS_DEFINES_H
#define ENGINE_CLIENT_GRAPHICS_DEFINES_H

#include <cstddef>
#include <cstdint>

typedef uint32_t TWGLuint;
typedef int32_t TWGLint;
typedef uint32_t TWGLenum;
typedef uint8_t TWGLubyte;

#endif
