/* AUTO GENERATED! DO NOT EDIT MANUALLY! See scripts/gen_keys.py */

#ifndef ENGINE_CLIENT_KEYNAMES_H
#define ENGINE_CLIENT_KEYNAMES_H

/**
 * Do not use directly! Use the @link IInput::KeyName @endlink function.
 */
extern const char g_aaKeyStrings[512][20];

#endif
