#pragma once
#include "base/rust.h"
#include "engine/console.h"
#include "engine/rust.h"
#include <cstdint>
#include <memory>

using IConsole_IResult = ::IConsole_IResult;
using IConsole = ::IConsole;
