/* (c) Magnus Auvinen. See licence.txt in the root of the distribution for more information. */
/* If you are missing that file, acquire a complete release at teeworlds.com.                */
#ifndef ENGINE_CLIENT_ENUMS_H
#define ENGINE_CLIENT_ENUMS_H

enum
{
	NUM_DUMMIES = 2,
};

#endif
